from decimal import Decimal
from django.db import models, transaction
from django.contrib.postgres.fields import HStoreField
from .settings import CCH_PRECISION


class OrderTaxation(models.Model):
    """
    Persist top-level taxation data related to an Order.
    """

    #: One-to-one foreign key to :class:`order.Order <oscar.apps.models.Order>`.
    order = models.OneToOneField('order.Order',
        related_name='taxation',
        on_delete=models.CASCADE,
        primary_key=True)

    #: Transaction ID returned by CCH
    transaction_id = models.IntegerField()

    #: Transaction Status returned by CCH
    transaction_status = models.IntegerField()

    #: Total Tax applied to the order
    total_tax_applied = models.DecimalField(decimal_places=2, max_digits=12)

    #: Message text returned by CCH
    messages = models.TextField(null=True)

    @classmethod
    def save_details(cls, order, taxes):
        """
        Given an order and a SOAP response, persist the details.

        :param order: :class:`Order <oscar.apps.order.models.Order>` instance
        :param taxes: Return value of :func:`CCHTaxCalculator.apply_taxes <oscarcch.calculator.CCHTaxCalculator.apply_taxes>`
        """
        with transaction.atomic():
            order_taxation = cls(order=order)
            order_taxation.transaction_id = taxes.TransactionID
            order_taxation.transaction_status = taxes.TransactionStatus
            order_taxation.total_tax_applied = Decimal(taxes.TotalTaxApplied).quantize(CCH_PRECISION)
            order_taxation.messages = taxes.Messages
            order_taxation.save()

            if taxes.LineItemTaxes:
                for cch_line in taxes.LineItemTaxes.LineItemTax:
                    line = order.lines.get(basket_line__id=cch_line.ID)
                    LineItemTaxation.save_details(line, cch_line)

    def __str__(self):
        return '%s' % (self.transaction_id)


class LineItemTaxation(models.Model):
    """
    Persist taxation details related to a single order line.
    """

    #: One-to-one foreign key to :class:`order.Line <oscar.apps.models.Line>`
    line_item = models.OneToOneField('order.Line',
        related_name='taxation',
        on_delete=models.CASCADE)

    #: Country code used to calculate taxes
    country_code = models.CharField(max_length=5)

    #: State code used to calculate taxes
    state_code = models.CharField(max_length=5)

    #: Total tax applied to the line
    total_tax_applied = models.DecimalField(decimal_places=2, max_digits=12)

    @classmethod
    def save_details(cls, line, taxes):
        with transaction.atomic():
            line_taxation = cls(line_item=line)
            line_taxation.country_code = taxes.CountryCode
            line_taxation.state_code = taxes.StateOrProvince
            line_taxation.total_tax_applied = Decimal(taxes.TotalTaxApplied).quantize(CCH_PRECISION)
            line_taxation.save()

            for detail in taxes.TaxDetails.TaxDetail:
                line_detail = LineItemTaxationDetail()
                line_detail.taxation = line_taxation
                line_detail.data = { str(k): str(v) for k, v in dict(detail).items() }
                line_detail.save()

    def __str__(self):
        return '%s: %s' % (self.line_item, self.total_tax_applied)


class LineItemTaxationDetail(models.Model):
    """
    Represents a single type tax applied to a line.
    """

    #: Many-to-one foreign key to :class:`LineItemTaxation <oscarcch.models.LineItemTaxation>`
    taxation = models.ForeignKey(LineItemTaxation,
        related_name='details',
        on_delete=models.CASCADE)

    #: HStore of data about the applied tax
    data = HStoreField()

    def __str__(self):
        return '%s—%s' % (self.data.get('AuthorityName'), self.data.get('TaxName'))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.contrib.postgres.fields.hstore
from django.contrib.postgres.operations import HStoreExtension


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        HStoreExtension(),
        migrations.CreateModel(
            name='LineItemTaxation',
            fields=[
                ('id', models.AutoField(primary_key=True, auto_created=True, serialize=False, verbose_name='ID')),
                ('country_code', models.CharField(max_length=5)),
                ('state_code', models.CharField(max_length=5)),
                ('total_tax_applied', models.DecimalField(max_digits=12, decimal_places=2)),
                ('line_item', models.OneToOneField(to='order.Line', related_name='taxation', on_delete=models.CASCADE)),
            ],
        ),
        migrations.CreateModel(
            name='LineItemTaxationDetail',
            fields=[
                ('id', models.AutoField(primary_key=True, auto_created=True, serialize=False, verbose_name='ID')),
                ('data', django.contrib.postgres.fields.hstore.HStoreField()),
                ('taxation', models.ForeignKey(to='cch.LineItemTaxation', related_name='details', on_delete=models.CASCADE)),
            ],
        ),
        migrations.CreateModel(
            name='OrderTaxation',
            fields=[
                ('order', models.OneToOneField(primary_key=True, serialize=False, to='order.Order', related_name='taxation', on_delete=models.CASCADE)),
                ('transaction_id', models.IntegerField()),
                ('transaction_status', models.IntegerField()),
                ('total_tax_applied', models.DecimalField(max_digits=12, decimal_places=2)),
                ('messages', models.TextField(null=True)),
            ],
        ),
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class CCHConfig(AppConfig):
    name = 'oscarcch'
    label = 'cch'
    # Translators: Backend Library Name
    verbose_name = _("CCH Sales Tax Office")

    def ready(self):
        from . import prices  # NOQA
        prices.monkey_patch_fixed_price()



from decimal import Decimal as D
from freezegun import freeze_time
from oscar.core.loading import get_model, get_class
from .base import BaseTest
from .base import p
import mock


Basket = get_model('basket', 'Basket')
ShippingAddress = get_model('order', 'ShippingAddress')
Country = get_model('address', 'Country')
PartnerAddress = get_model('partner', 'PartnerAddress')
Range = get_model('offer', 'Range')
Benefit = get_model('offer', 'Benefit')
Condition = get_model('offer', 'Condition')
ConditionalOffer = get_model('offer', 'ConditionalOffer')

USStrategy = get_class('partner.strategy', 'US')
Applicator = get_class('offer.applicator', 'Applicator')
CCHTaxCalculator = get_class('oscarcch.calculator', 'CCHTaxCalculator')


class CCHTaxCalculatorRealTest(BaseTest):

    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.get_transport')
    def test_apply_taxes_five_digits_postal_code(self, get_transport ):
        basket = self.prepare_basket_full_zip()
        to_address = self.get_to_address_ohio_short_zip()

        def test_request(request):
            self.assertNodeText(request.message, p('Body/CalculateRequest/EntityID'), 'TESTSANDBOX')
            self.assertNodeText(request.message, p('Body/CalculateRequest/DivisionID'), '42')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/CustomerType'), '08')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/InvoiceDate'), '2016-04-13T12:14:44.018599-04:00')
            self.assertNodeCount(request.message, p('Body/CalculateRequest/order/LineItems/LineItem'), 1)
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/AvgUnitPrice'), '10.00000')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ID'), str(basket.all_lines()[0].id))
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/City'), 'Anchorage')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line1'), '325 F st')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/PostalCode'), '99501')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Plus4'), '2217')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/StateOrProvince'), 'AK')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/City'), 'BRINKHAVEN')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line1'), '33001 STATE ROUTE 206')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/PostalCode'), '43006')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/StateOrProvince'), 'OH')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/Quantity'), '1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/SKU'), 'ABC123')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/ProviderType'), '70')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/SourceSystem'), 'Oscar')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TestTransaction'), 'true')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionID'), '0')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionType'), '01')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/finalize'), 'false')
        resp = self._get_cch_response_ohio_request_short_zip( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))


        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.68'))
        self.assertEqual(basket.total_tax, D('0.68'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.68'))
        self.assertEqual(purchase_info.price.tax, D('0.68'))

        details = purchase_info.price.taxation_details
        self.assertEqual(len(details), 2)
        self.assertEqual(details[0].authority_name, 'OHIO, STATE OF')
        self.assertEqual(details[0].tax_name, 'STATE SALES TAX-GENERAL MERCHANDISE')
        self.assertEqual(details[0].tax_applied, D('0.58'))
        self.assertEqual(details[0].fee_applied, D('0.00'))

    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.get_transport')
    def test_apply_taxes_nine_digits_postal_code(self, get_transport):
        basket = self.prepare_basket_full_zip()
        to_address = self.get_to_address_ohio_full_zip()

        def test_request(request):
            self.assertNodeText(request.message, p('Body/CalculateRequest/EntityID'), 'TESTSANDBOX')
            self.assertNodeText(request.message, p('Body/CalculateRequest/DivisionID'), '42')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/CustomerType'), '08')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/InvoiceDate'), '2016-04-13T12:14:44.018599-04:00')
            self.assertNodeCount(request.message, p('Body/CalculateRequest/order/LineItems/LineItem'), 1)
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/AvgUnitPrice'), '10.00000')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ID'), str(basket.all_lines()[0].id))
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/City'), 'Anchorage')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line1'), '325 F st')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/PostalCode'), '99501')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Plus4'), '2217')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/StateOrProvince'), 'AK')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/City'), 'BRINKHAVEN')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line1'), '200 HIGH ST')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/PostalCode'), '43006')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Plus4'), '9000')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/StateOrProvince'), 'OH')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/Quantity'), '1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/SKU'), 'ABC123')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/ProviderType'), '70')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/SourceSystem'), 'Oscar')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TestTransaction'), 'true')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionID'), '0')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionType'), '01')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/finalize'), 'false')
        resp = self._get_cch_response_ohio_request_full_zip( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.73'))
        self.assertEqual(basket.total_tax, D('0.73'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.73'))
        self.assertEqual(purchase_info.price.tax, D('0.73'))

        details = purchase_info.price.taxation_details
        self.assertEqual(len(details), 2)
        self.assertEqual(details[0].authority_name, 'OHIO, STATE OF')
        self.assertEqual(details[0].tax_name, 'STATE SALES TAX-GENERAL MERCHANDISE')
        self.assertEqual(details[0].tax_applied, D('0.58'))
        self.assertEqual(details[0].fee_applied, D('0.00'))



from django.test import TestCase
from oscar.apps.partner import prices
from oscar.core.prices import TaxNotKnown
from oscar.core.loading import get_model, get_class
from decimal import Decimal as D

Basket = get_model('basket', 'Basket')
BasketLine = get_model('basket', 'Line')
ShippingAddress = get_model('order', 'ShippingAddress')
Country = get_model('address', 'Country')
USStrategy = get_class('partner.strategy', 'US')


class FixedPriceTaxationDetailsTest(TestCase):
    def test_add_taxes(self):
        p = prices.FixedPrice('USD', D('10.00'))

        def get_incl_tax():
            return p.incl_tax

        self.assertEqual(p.excl_tax, D('10.00'))
        self.assertFalse(p.is_tax_known)
        self.assertRaises(TaxNotKnown, get_incl_tax)
        self.assertIsNone(p.tax)

        p.add_tax('New York State', 'Sales Tax', tax_applied='1.00')

        self.assertEqual(p.excl_tax, D('10.00'))
        self.assertTrue(p.is_tax_known)
        self.assertEqual(p.incl_tax, D('11.00'))
        self.assertEqual(p.tax, D('1.00'))

        p.add_tax('IRS', 'Some Fee', fee_applied='0.15')

        self.assertEqual(p.excl_tax, D('10.00'))
        self.assertTrue(p.is_tax_known)
        self.assertEqual(p.incl_tax, D('11.15'))
        self.assertEqual(p.tax, D('1.15'))

        self.assertEqual(p.taxation_details[0].authority_name, 'New York State')
        self.assertEqual(p.taxation_details[0].tax_name, 'Sales Tax')
        self.assertEqual(p.taxation_details[0].tax_applied, D('1.00'))
        self.assertEqual(p.taxation_details[0].fee_applied, D('0.00'))

        self.assertEqual(p.taxation_details[1].authority_name, 'IRS')
        self.assertEqual(p.taxation_details[1].tax_name, 'Some Fee')
        self.assertEqual(p.taxation_details[1].tax_applied, D('0.00'))
        self.assertEqual(p.taxation_details[1].fee_applied, D('0.15'))

        p.clear_taxes()

        self.assertEqual(p.excl_tax, D('10.00'))
        self.assertFalse(p.is_tax_known)
        self.assertRaises(TaxNotKnown, get_incl_tax)
        self.assertIsNone(p.tax)



from django.contrib import admin
from oscar.core.loading import get_model

OrderTaxation = get_model('cch', 'OrderTaxation')
LineItemTaxationDetail = get_model('cch', 'LineItemTaxationDetail')
LineItemTaxation = get_model('cch', 'LineItemTaxation')


@admin.register(OrderTaxation)
class OrderTaxationAdmin(admin.ModelAdmin):
    list_filter = ['transaction_status']
    search_fields = ['transaction_id', 'messages']

    fields = ['order', 'transaction_id', 'transaction_status', 'total_tax_applied', 'messages']
    list_display = fields
    readonly_fields = fields


class DetailInline(admin.StackedInline):
    model = LineItemTaxationDetail
    readonly_fields = ['data']


@admin.register(LineItemTaxation)
class LineItemTaxationAdmin(admin.ModelAdmin):
    list_filter = ['country_code', 'state_code']

    fields = ['line_item', 'country_code', 'state_code', 'total_tax_applied']
    list_display = fields
    readonly_fields = fields
    inlines = [DetailInline]




class CCHError(Exception):
    severity = None
    code = None
    info = None

    def __init__(self, code, info, *args, **kwargs):
        self.code = code
        self.info = info
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.message


class CCHSystemError(CCHError):
    severity = 1

    @property
    def message(self):
        return 'CCHSystemError %s: %s' % (self.code, self.info)


class CCHRequestError(CCHError):
    severity = 2

    @property
    def message(self):
        return 'CCHRequestError %s: %s' % (self.code, self.info)


def build(severity, code, info):
    types = {
        CCHSystemError.severity: CCHSystemError,
        CCHRequestError.severity: CCHRequestError,
    }
    Exc = types.get(severity, CCHError)
    return Exc(code, info)



from decimal import Decimal as D
from oscar.core.loading import get_model, get_class
from oscar.test import factories
from .base import BaseTest
import mock

Basket = get_model('basket', 'Basket')
ShippingAddress = get_model('order', 'ShippingAddress')
Country = get_model('address', 'Country')
USStrategy = get_class('partner.strategy', 'US')


class PersistCCHDetailsTest(BaseTest):
    @mock.patch('soap.get_transport')
    def test_persist_taxation_details(self, get_transport):
        """Place an order with normal taxes"""
        basket = self.prepare_basket()
        to_address = self.get_to_address()

        resp = self._get_cch_response_normal( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp)

        # This should call CCH and save tax details in the DB
        order = factories.create_order(basket=basket, shipping_address=to_address)

        # Make sure we have an order taxation object
        self.assertEqual(order.taxation.transaction_id, 40043)
        self.assertEqual(order.taxation.transaction_status, 4)
        self.assertEqual(order.taxation.total_tax_applied, D('0.89'))

        # Make sure we have an line taxation objects
        for line in order.lines.all():
            self.assertEqual(line.taxation.country_code, 'US')
            self.assertEqual(line.taxation.state_code, 'NY')
            self.assertEqual(line.taxation.total_tax_applied, D('0.89'))
            self.assertEqual(line.taxation.details.count(), 3)
            for detail in line.taxation.details.all():
                self.assertIn('AuthorityName', detail.data)
                self.assertIn('TaxName', detail.data)
                self.assertIn('TaxRate', detail.data)
                self.assertTrue(float(detail.data['TaxRate']) > 0)
                self.assertIn('TaxableAmount', detail.data)
                self.assertIn('TaxableQuantity', detail.data)

    @mock.patch('soap.get_transport')
    def test_persist_taxation_details_when_zero(self, get_transport):
        """Place a tax free order"""
        basket = self.prepare_basket()
        to_address = self.get_to_address()

        resp = self._get_cch_response_empty()
        get_transport.return_value = self._build_transport_with_reply(resp)

        # This should call CCH and save tax details in the DB
        order = factories.create_order(basket=basket, shipping_address=to_address)

        # Make sure we have an order taxation object
        self.assertEqual(order.taxation.transaction_id, 40043)
        self.assertEqual(order.taxation.transaction_status, 4)
        self.assertEqual(order.taxation.total_tax_applied, D('0.00'))

        # Make sure we have an line taxation objects
        for line in order.lines.all():
            self.assertFalse(hasattr(line, 'taxation'))



from datetime import datetime
from decimal import Decimal
from django_statsd.clients import statsd
from . import exceptions, settings
import logging
import requests
import soap

logger = logging.getLogger(__name__)

POSTCODE_LEN = 5
PLUS4_LEN = 4


class CCHTaxCalculator(object):
    """
    Simple interface between Python and the CCH Sales Tax Office SOAP API.
    """
    precision = settings.CCH_PRECISION
    wsdl = settings.CCH_WSDL
    entity_id = settings.CCH_ENTITY
    divsion_id = settings.CCH_DIVISION
    max_retries = settings.CCH_MAX_RETRIES


    def estimate_taxes(self, basket, shipping_address):
        """
        DEPRECATED. Use :func:`CCHTaxCalculator.apply_taxes <oscarcch.calculator.CCHTaxCalculator.apply_taxes>` instead.
        """
        statsd.incr('cch.estimate')
        self.apply_taxes(basket, shipping_address)
        return basket


    def apply_taxes(self, basket, shipping_address, ignore_cch_fail=False):
        """
        Apply taxes to a Basket instance using the given shipping address.

        Pass return value of this method to :func:`OrderTaxation.save_details <oscarcch.models.OrderTaxation.save_details>`
        to persist the taxation details, CCH transaction ID, etc in the database.

        :param basket: :class:`Basket <oscar.apps.basket.models.Basket>` instance
        :param shipping_address: :class:`ShippingAddress <oscar.apps.order.models.ShippingAddress>` instance
        :param ignore_cch_fail: When `True`, allows CCH to fail silently
        :return: SOAP Response.
        """
        with statsd.timer('cch.apply-time'):
            response = self._get_response(basket, shipping_address, ignore_cch_fail)

        if not ignore_cch_fail:
            self._check_response_messages(response)

        # Apply taxes to line items
        for line in basket.all_lines():
            line_id = str(line.id)

            taxes = None
            if response and response.LineItemTaxes:
                try:
                    taxes = next(filter(lambda item: item.ID == line_id, response.LineItemTaxes.LineItemTax))
                except StopIteration:
                    pass

            # Taxes come in two forms: quantity and percentage based
            # We need to handle both of those here. The tricky part is that CCH returns data
            # for an entire line item (inclusive quantity), but Oscar needs the tax info for
            # each unit in the line (exclusive quantity). So, we use the details provided to
            # derive the per-unit taxes before applying them.
            line.purchase_info.price.clear_taxes()
            if taxes:
                for tax in taxes.TaxDetails.TaxDetail:
                    unit_fee = Decimal(str(tax.FeeApplied)) / line.quantity
                    unit_tax = Decimal(str(tax.TaxApplied)) / line.quantity
                    line.purchase_info.price.add_tax(
                        authority_name=tax.AuthorityName,
                        tax_name=tax.TaxName,
                        tax_applied=unit_tax,
                        fee_applied=unit_fee)
                # Check our work and make sure the total we arrived at matches the total CCH gave us
                total_line_tax = (line.purchase_info.price.tax * line.quantity).quantize(self.precision)
                total_applied_tax = Decimal(taxes.TotalTaxApplied).quantize(self.precision)
                if total_applied_tax != total_line_tax:
                    statsd.incr('cch.miscalculation')
                    raise RuntimeError((
                        "Taxation miscalculation occurred! "
                        "Details sum to %s, which doesn't match given sum of %s"
                    ) % (total_line_tax, taxes.TotalTaxApplied))
            else:
                line.purchase_info.price.tax = Decimal('0.00')

        return response


    def _get_response(self, basket, shipping_address, ignore_cch_fail=False):
        """Fetch CCH tax data for the given basket and shipping address"""
        response = None
        retry_count = 0
        while response is None and retry_count <= self.max_retries:
            response = self._get_response_inner(basket, shipping_address, ignore_cch_fail, retry_count=retry_count)
            retry_count += 1

        if response:
            statsd.incr('cch.apply-success')
        else:
            statsd.incr('cch.apply-failure')

        return response


    def _get_response_inner(self, basket, shipping_address, ignore_cch_fail, retry_count):
        # Attempt to get a response
        response = None
        try:
            order = self._build_order(basket, shipping_address)
            response = self.client.service.CalculateRequest(self.entity_id, self.divsion_id, order)

        # Timeouts (read or connect) get retried
        except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
            if retry_count >= self.max_retries and not ignore_cch_fail:
                raise e

        # Catch any other possible exceptions and, if we're supposed to, ignore them.
        except Exception as e:
            if not ignore_cch_fail:
                raise e

        # Return our (apparently successful) response
        return response


    def _check_response_messages(self, response):
        """Raise an exception if response messages contains any reported errors."""
        if response and response.Messages:
            for message in response.Messages.Message:
                if message.Code > 0:
                    raise exceptions.build(message.Severity, message.Code, message.Info)


    @property
    def client(self):
        """Lazy constructor for SOAP client"""
        return soap.get_client(self.wsdl, 'CCH')


    def _build_order(self, basket, shipping_address):
        """Convert an Oscar Basket and ShippingAddresss into a CCH Order object"""
        order = self.client.factory.create('ns15:Order')

        order.InvoiceDate = datetime.now(settings.CCH_TIME_ZONE)
        order.SourceSystem = settings.CCH_SOURCE_SYSTEM
        order.TestTransaction = settings.CCH_TEST_TRANSACTIONS
        order.TransactionType = settings.CCH_TRANSACTION_TYPE
        order.CustomerType = settings.CCH_CUSTOMER_TYPE
        order.ProviderType = settings.CCH_PROVIDER_TYPE
        order.TransactionID = 0
        order.finalize = settings.CCH_FINALIZE_TRANSACTION

        for line in basket.all_lines():
            qty = getattr(line, 'cch_quantity', line.quantity)
            if qty <= 0:
                continue

            item = self.client.factory.create('ns11:LineItem')
            item.ID = line.id
            item.AvgUnitPrice = (line.line_price_excl_tax_incl_discounts / qty).quantize(Decimal('0.00001'))
            item.Quantity = qty
            item.ExemptionCode = None
            item.SKU = self._get_product_data('sku', line)
            item.ProductInfo = self.client.factory.create('ns21:ProductInfo')
            item.ProductInfo.ProductGroup = self._get_product_data('group', line)
            item.ProductInfo.ProductItem = self._get_product_data('item', line)

            item.NexusInfo = self.client.factory.create('ns14:NexusInfo')

            item.NexusInfo.ShipFromAddress = self.client.factory.create('ns0:Address')
            warehouse = line.stockrecord.partner.primary_address
            if warehouse:
                item.NexusInfo.ShipFromAddress.Line1 = warehouse.line1
                item.NexusInfo.ShipFromAddress.Line2 = warehouse.line2
                item.NexusInfo.ShipFromAddress.City = warehouse.city
                item.NexusInfo.ShipFromAddress.StateOrProvince = warehouse.state
                postcode, plus4 = self.format_postcode(warehouse.postcode)
                item.NexusInfo.ShipFromAddress.PostalCode = postcode
                item.NexusInfo.ShipFromAddress.Plus4 = plus4
                item.NexusInfo.ShipFromAddress.CountryCode = warehouse.country.code
            item.NexusInfo.ShipToAddress = self.client.factory.create('ns0:Address')
            item.NexusInfo.ShipToAddress.Line1 = shipping_address.line1
            item.NexusInfo.ShipToAddress.Line2 = shipping_address.line2
            item.NexusInfo.ShipToAddress.City = shipping_address.city
            item.NexusInfo.ShipToAddress.StateOrProvince = shipping_address.state
            postcode, plus4 = self.format_postcode(shipping_address.postcode)
            item.NexusInfo.ShipToAddress.PostalCode = postcode
            item.NexusInfo.ShipToAddress.Plus4 = plus4
            item.NexusInfo.ShipToAddress.CountryCode = shipping_address.country.code

            order.LineItems.LineItem.append(item)

        return order


    def _get_product_data(self, key, line):
        key = 'cch_product_%s' % key
        sku = getattr(settings, key.upper())
        sku = getattr(line.product.attr, key.lower(), sku)
        return sku

    def format_postcode(self, raw_postcode):
        postcode, plus4 = raw_postcode[:POSTCODE_LEN], None
        # set Plus4 if PostalCode provided as 9 digits separated by hyphen
        if len(raw_postcode) == POSTCODE_LEN + PLUS4_LEN + 1:
            plus4 = raw_postcode[POSTCODE_LEN + 1:]
        return postcode, plus4



from decimal import Decimal as D
from freezegun import freeze_time
from oscar.core.loading import get_model, get_class
from oscar.test import factories
from .base import BaseTest
from .base import p
import mock
import requests

Basket = get_model('basket', 'Basket')
ShippingAddress = get_model('order', 'ShippingAddress')
Country = get_model('address', 'Country')
PartnerAddress = get_model('partner', 'PartnerAddress')
Range = get_model('offer', 'Range')
Benefit = get_model('offer', 'Benefit')
Condition = get_model('offer', 'Condition')
ConditionalOffer = get_model('offer', 'ConditionalOffer')

USStrategy = get_class('partner.strategy', 'US')
Applicator = get_class('offer.applicator', 'Applicator')
CCHTaxCalculator = get_class('oscarcch.calculator', 'CCHTaxCalculator')


class CCHTaxCalculatorTest(BaseTest):

    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.get_transport')
    def test_apply_taxes_normal(self, get_transport):
        basket = self.prepare_basket()
        to_address = self.get_to_address()

        def test_request(request):
            self.assertNodeText(request.message, p('Body/CalculateRequest/EntityID'), 'TESTSANDBOX')
            self.assertNodeText(request.message, p('Body/CalculateRequest/DivisionID'), '42')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/CustomerType'), '08')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/InvoiceDate'), '2016-04-13T12:14:44.018599-04:00')
            self.assertNodeCount(request.message, p('Body/CalculateRequest/order/LineItems/LineItem'), 1)
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/AvgUnitPrice'), '10.00000')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ID'), str(basket.all_lines()[0].id))
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/City'), 'Anchorage')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line1'), '221 Baker st')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line2'), 'B')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/PostalCode'), '99501')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/StateOrProvince'), 'AK')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/City'), 'Brooklyn')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line1'), '123 Evergreen Terrace')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line2'), 'Apt #1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/PostalCode'), '11201')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/StateOrProvince'), 'NY')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/Quantity'), '1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/SKU'), 'ABC123')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/ProviderType'), '70')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/SourceSystem'), 'Oscar')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TestTransaction'), 'true')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionID'), '0')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionType'), '01')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/finalize'), 'false')
        resp = self._get_cch_response_normal( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.89'))
        self.assertEqual(basket.total_tax, D('0.89'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.89'))
        self.assertEqual(purchase_info.price.tax, D('0.89'))

        details = purchase_info.price.taxation_details
        self.assertEqual(len(details), 3)
        self.assertEqual(details[0].authority_name, 'NEW YORK, STATE OF')
        self.assertEqual(details[0].tax_name, 'STATE SALES TAX-GENERAL MERCHANDISE')
        self.assertEqual(details[0].tax_applied, D('0.40'))
        self.assertEqual(details[0].fee_applied, D('0.00'))


    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.http.requests')
    def test_apply_taxes_read_timeout(self, soap_requests):
        basket = self.prepare_basket()
        to_address = self.get_to_address()

        # Make requests throw a ReadTimeout
        def raise_error(*args, **kwargs):
            raise requests.exceptions.ReadTimeout()
        soap_requests.post = mock.MagicMock(side_effect=raise_error)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        with self.assertRaises(requests.exceptions.ReadTimeout):
            CCHTaxCalculator().apply_taxes(basket, to_address)


    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.http.requests')
    def test_apply_taxes_read_timeout_retried(self, soap_requests):
        basket = self.prepare_basket()
        to_address = self.get_to_address()
        closured = {'i': 0}

        # Make requests throw a ReadTimeout, but only the first time.
        def raise_error(*args, **kwargs):
            if closured['i'] > 0:
                resp = mock.MagicMock()
                resp.headers = {}
                resp.content = self._get_cch_response_normal( basket.all_lines()[0].id )
                return resp
            closured['i'] += 1
            raise requests.exceptions.ReadTimeout()
        soap_requests.post = mock.MagicMock(side_effect=raise_error)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertEqual(soap_requests.post.call_count, 2)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.89'))
        self.assertEqual(basket.total_tax, D('0.89'))


    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.get_transport')
    def test_apply_taxes_after_discount(self, get_transport):
        basket = self.prepare_basket()
        to_address = self.get_to_address()

        # Make an offer to apply to the basket
        rng = Range.objects.create(name='Some products', includes_all_products=True)
        condition = Condition.objects.create(range=rng, type=Condition.VALUE, value=D('0.01'))
        benefit = Benefit.objects.create(range=rng, type=Benefit.PERCENTAGE, value=D('50.00'))
        ConditionalOffer.objects.create(name='My Discount', condition=condition, benefit=benefit)

        # Apply offers to basket
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        Applicator().apply(basket)
        self.assertEqual(basket.total_excl_tax, D('5.00'))

        def test_request(request):
            self.assertNodeText(request.message, p('Body/CalculateRequest/EntityID'), 'TESTSANDBOX')
            self.assertNodeText(request.message, p('Body/CalculateRequest/DivisionID'), '42')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/CustomerType'), '08')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/InvoiceDate'), '2016-04-13T12:14:44.018599-04:00')
            self.assertNodeCount(request.message, p('Body/CalculateRequest/order/LineItems/LineItem'), 1)
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/AvgUnitPrice'), '5.00000')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ID'), str(basket.all_lines()[0].id))
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/City'), 'Anchorage')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line1'), '221 Baker st')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line2'), 'B')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/PostalCode'), '99501')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/StateOrProvince'), 'AK')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/City'), 'Brooklyn')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line1'), '123 Evergreen Terrace')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line2'), 'Apt #1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/PostalCode'), '11201')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/StateOrProvince'), 'NY')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/Quantity'), '1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/SKU'), 'ABC123')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/ProviderType'), '70')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/SourceSystem'), 'Oscar')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TestTransaction'), 'true')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionID'), '0')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionType'), '01')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/finalize'), 'false')
        resp = self._get_cch_response_normal( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('5.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('5.00'))
        self.assertEqual(basket.total_incl_tax, D('5.89'))
        self.assertEqual(basket.total_tax, D('0.89'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.89'))
        self.assertEqual(purchase_info.price.tax, D('0.89'))

        details = purchase_info.price.taxation_details
        self.assertEqual(len(details), 3)
        self.assertEqual(details[0].authority_name, 'NEW YORK, STATE OF')
        self.assertEqual(details[0].tax_name, 'STATE SALES TAX-GENERAL MERCHANDISE')
        self.assertEqual(details[0].tax_applied, D('0.40'))
        self.assertEqual(details[0].fee_applied, D('0.00'))


    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.get_transport')
    def test_custom_quantity(self, get_transport):
        basket = self.prepare_basket()
        to_address = self.get_to_address()

        # Set a custom cch_quantity property
        for basket_line in basket.all_lines():
            basket_line.cch_quantity = 3

        def test_request(request):
            self.assertNodeCount(request.message, p('Body/CalculateRequest/order/LineItems/LineItem'), 1)
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/AvgUnitPrice'), '3.33333')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ID'), str(basket_line.id))
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/Quantity'), '3')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/SKU'), 'ABC123')

        resp = self._get_cch_response_normal(basket_line.id)
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.89'))
        self.assertEqual(basket.total_tax, D('0.89'))


    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.get_transport')
    def test_apply_taxes_custom_sku(self, get_transport):
        basket = Basket()
        basket.strategy = USStrategy()

        product = factories.create_product(attributes={
            'cch_product_sku': 'CCH Product SKU',
        })
        product.attr.cch_product_sku = 'XYZ456'
        product.save()
        record = factories.create_stockrecord(
            currency='USD',
            product=product,
            price_excl_tax=D('10.00'))
        factories.create_purchase_info(record)
        basket.add(product)

        from_address = PartnerAddress()
        from_address.line1 = '221 Baker st'
        from_address.line2 = 'B'
        from_address.line4 = 'Anchorage'
        from_address.state = 'AK'
        from_address.postcode = '99501'
        from_address.country = Country.objects.get(pk='US')
        from_address.partner = record.partner
        from_address.save()

        to_address = self.get_to_address()

        def test_request(request):
            self.assertNodeText(request.message, p('Body/CalculateRequest/EntityID'), 'TESTSANDBOX')
            self.assertNodeText(request.message, p('Body/CalculateRequest/DivisionID'), '42')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/CustomerType'), '08')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/InvoiceDate'), '2016-04-13T12:14:44.018599-04:00')
            self.assertNodeCount(request.message, p('Body/CalculateRequest/order/LineItems/LineItem'), 1)
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/AvgUnitPrice'), '10.00000')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ID'), str(basket.all_lines()[0].id))
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/City'), 'Anchorage')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line1'), '221 Baker st')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line2'), 'B')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/PostalCode'), '99501')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/StateOrProvince'), 'AK')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/City'), 'Brooklyn')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line1'), '123 Evergreen Terrace')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line2'), 'Apt #1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/PostalCode'), '11201')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/StateOrProvince'), 'NY')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/Quantity'), '1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/SKU'), 'XYZ456')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/ProviderType'), '70')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/SourceSystem'), 'Oscar')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TestTransaction'), 'true')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionID'), '0')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionType'), '01')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/finalize'), 'false')
        resp = self._get_cch_response_normal( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.89'))
        self.assertEqual(basket.total_tax, D('0.89'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.89'))
        self.assertEqual(purchase_info.price.tax, D('0.89'))


    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.get_transport')
    def test_apply_taxes_custom_sku_empty(self, get_transport):
        basket = Basket()
        basket.strategy = USStrategy()

        product = factories.create_product(attributes={
            'cch_product_sku': 'CCH Product SKU',
        })
        product.attr.cch_product_sku = ''
        product.save()
        record = factories.create_stockrecord(
            currency='USD',
            product=product,
            price_excl_tax=D('10.00'))
        factories.create_purchase_info(record)
        basket.add(product)

        from_address = PartnerAddress()
        from_address.line1 = '221 Baker st'
        from_address.line2 = 'B'
        from_address.line4 = 'Anchorage'
        from_address.state = 'AK'
        from_address.postcode = '99501'
        from_address.country = Country.objects.get(pk='US')
        from_address.partner = record.partner
        from_address.save()

        to_address = self.get_to_address()

        def test_request(request):
            self.assertNodeText(request.message, p('Body/CalculateRequest/EntityID'), 'TESTSANDBOX')
            self.assertNodeText(request.message, p('Body/CalculateRequest/DivisionID'), '42')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/CustomerType'), '08')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/InvoiceDate'), '2016-04-13T12:14:44.018599-04:00')
            self.assertNodeCount(request.message, p('Body/CalculateRequest/order/LineItems/LineItem'), 1)
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/AvgUnitPrice'), '10.00000')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ID'), str(basket.all_lines()[0].id))
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/City'), 'Anchorage')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line1'), '221 Baker st')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line2'), 'B')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/PostalCode'), '99501')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/StateOrProvince'), 'AK')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/City'), 'Brooklyn')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line1'), '123 Evergreen Terrace')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line2'), 'Apt #1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/PostalCode'), '11201')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/StateOrProvince'), 'NY')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/Quantity'), '1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/SKU'), 'ABC123')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/ProviderType'), '70')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/SourceSystem'), 'Oscar')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TestTransaction'), 'true')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionID'), '0')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionType'), '01')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/finalize'), 'false')
        resp = self._get_cch_response_normal( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.89'))
        self.assertEqual(basket.total_tax, D('0.89'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.89'))
        self.assertEqual(purchase_info.price.tax, D('0.89'))


    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.get_transport')
    def test_apply_taxes_custom_product_data(self, get_transport):
        basket = Basket()
        basket.strategy = USStrategy()

        product = factories.create_product(attributes={
            'cch_product_group': 'CCH Product Item',
            'cch_product_item': 'CCH Product Group',
        })
        product.attr.cch_product_group = 'MyGroup'
        product.attr.cch_product_item = 'MyItem'
        product.save()
        record = factories.create_stockrecord(
            currency='USD',
            product=product,
            price_excl_tax=D('10.00'))
        factories.create_purchase_info(record)
        basket.add(product)

        from_address = PartnerAddress()
        from_address.line1 = '221 Baker st'
        from_address.line2 = 'B'
        from_address.line4 = 'Anchorage'
        from_address.state = 'AK'
        from_address.postcode = '99501'
        from_address.country = Country.objects.get(pk='US')
        from_address.partner = record.partner
        from_address.save()

        to_address = self.get_to_address()

        def test_request(request):
            self.assertNodeText(request.message, p('Body/CalculateRequest/EntityID'), 'TESTSANDBOX')
            self.assertNodeText(request.message, p('Body/CalculateRequest/DivisionID'), '42')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/CustomerType'), '08')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/InvoiceDate'), '2016-04-13T12:14:44.018599-04:00')
            self.assertNodeCount(request.message, p('Body/CalculateRequest/order/LineItems/LineItem'), 1)
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/AvgUnitPrice'), '10.00000')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ID'), str(basket.all_lines()[0].id))
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/City'), 'Anchorage')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line1'), '221 Baker st')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/Line2'), 'B')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/PostalCode'), '99501')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/StateOrProvince'), 'AK')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/City'), 'Brooklyn')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/CountryCode'), 'US')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line1'), '123 Evergreen Terrace')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/Line2'), 'Apt #1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/PostalCode'), '11201')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/StateOrProvince'), 'NY')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/Quantity'), '1')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/SKU'), 'ABC123')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ProductInfo/ProductGroup'), 'MyGroup')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/ProductInfo/ProductItem'), 'MyItem')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/ProviderType'), '70')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/SourceSystem'), 'Oscar')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TestTransaction'), 'true')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionID'), '0')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/TransactionType'), '01')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/finalize'), 'false')
        resp = self._get_cch_response_normal( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.89'))
        self.assertEqual(basket.total_tax, D('0.89'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.89'))
        self.assertEqual(purchase_info.price.tax, D('0.89'))


    @mock.patch('soap.get_transport')
    def test_truncate_postal_code(self, get_transport):
        basket = self.prepare_basket()
        to_address = self.get_to_address()

        def test_request(request):
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipFromAddress/PostalCode'), '99501')
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/NexusInfo/ShipToAddress/PostalCode'), '11201')
        resp = self._get_cch_response_normal( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        from_address = basket.all_lines()[0].stockrecord.partner.primary_address
        from_address.postcode = '99501-1234'
        from_address.save()

        to_address.postcode = '11201-9876'
        to_address.save()

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.89'))


    @mock.patch('soap.get_transport')
    def test_apply_taxes_repeatedly(self, get_transport):
        basket = self.prepare_basket()
        to_address = self.get_to_address()
        resp = self._get_cch_response_normal( basket.all_lines()[0].id )
        get_transport.return_value = self._build_transport_with_reply(resp)

        def assert_taxes_are_correct():
            self.assertTrue(basket.is_tax_known)
            self.assertEqual(basket.total_excl_tax, D('10.00'))
            self.assertEqual(basket.total_incl_tax, D('10.89'))
            self.assertEqual(basket.total_tax, D('0.89'))

            purchase_info = basket.all_lines()[0].purchase_info
            self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
            self.assertEqual(purchase_info.price.incl_tax, D('10.89'))
            self.assertEqual(purchase_info.price.tax, D('0.89'))

            details = purchase_info.price.taxation_details
            self.assertEqual(len(details), 3)
            self.assertEqual(details[0].authority_name, 'NEW YORK, STATE OF')
            self.assertEqual(details[0].tax_name, 'STATE SALES TAX-GENERAL MERCHANDISE')
            self.assertEqual(details[0].tax_applied, D('0.40'))
            self.assertEqual(details[0].fee_applied, D('0.00'))

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)
        assert_taxes_are_correct()

        CCHTaxCalculator().apply_taxes(basket, to_address)
        assert_taxes_are_correct()

        CCHTaxCalculator().apply_taxes(basket, to_address)
        assert_taxes_are_correct()


    @mock.patch('soap.get_transport')
    def test_apply_taxes_tax_free(self, get_transport):
        basket = self.prepare_basket()
        to_address = self.get_to_address()

        resp = self._get_cch_response_empty()
        get_transport.return_value = self._build_transport_with_reply(resp)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.00'))
        self.assertEqual(basket.total_tax, D('0.00'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.tax, D('0.00'))

        details = purchase_info.price.taxation_details
        self.assertEqual(len(details), 0)


    @mock.patch('soap.get_transport')
    def test_estimate_taxes(self, get_transport):
        basket = self.prepare_basket()
        to_address = self.get_to_address()
        resp = self._get_cch_response_normal( basket.all_lines()[0].id )
        transport = self._build_transport_with_reply(resp)
        get_transport.return_value = transport

        def assert_taxes_are_correct(b):
            self.assertTrue(b.is_tax_known)
            self.assertEqual(b.total_excl_tax, D('10.00'))
            self.assertEqual(b.total_incl_tax, D('10.89'))
            self.assertEqual(b.total_tax, D('0.89'))

            purchase_info = b.all_lines()[0].purchase_info
            self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
            self.assertEqual(purchase_info.price.incl_tax, D('10.89'))
            self.assertEqual(purchase_info.price.tax, D('0.89'))

            details = purchase_info.price.taxation_details
            self.assertEqual(len(details), 3)
            self.assertEqual(details[0].authority_name, 'NEW YORK, STATE OF')
            self.assertEqual(details[0].tax_name, 'STATE SALES TAX-GENERAL MERCHANDISE')
            self.assertEqual(details[0].tax_applied, D('0.40'))
            self.assertEqual(details[0].fee_applied, D('0.00'))

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        self.assertEqual(transport.send.call_count, 0)
        b1 = CCHTaxCalculator().estimate_taxes(basket, to_address)
        self.assertEqual(transport.send.call_count, 1)
        assert_taxes_are_correct(b1)

        b2 = CCHTaxCalculator().estimate_taxes(basket, to_address)
        self.assertEqual(transport.send.call_count, 2)
        assert_taxes_are_correct(b2)

        b3 = CCHTaxCalculator().estimate_taxes(basket, to_address)
        self.assertEqual(transport.send.call_count, 3)
        assert_taxes_are_correct(b3)

        basket.save()

        b4 = CCHTaxCalculator().estimate_taxes(basket, to_address)
        self.assertEqual(transport.send.call_count, 4)
        assert_taxes_are_correct(b4)


    @freeze_time("2016-04-13T16:14:44.018599-00:00")
    @mock.patch('soap.get_transport')
    def test_apply_taxes_zero_qty_line(self, get_transport):
        basket = self.prepare_basket(lines=2)
        to_address = self.get_to_address()

        basket.add_product(basket.all_lines()[0].product, -1)

        def test_request(request):
            self.assertNodeCount(request.message, p('Body/CalculateRequest/order/LineItems/LineItem'), 1)
            self.assertNodeText(request.message, p('Body/CalculateRequest/order/LineItems/LineItem/AvgUnitPrice'), '10.00000')

        resp = self._get_cch_response_normal(basket.all_lines()[1].id)
        get_transport.return_value = self._build_transport_with_reply(resp, test_request=test_request)

        self.assertFalse(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.89'))
        self.assertEqual(basket.total_tax, D('0.89'))

        self.assertEqual(len(basket.all_lines()), 2)



from collections import namedtuple
from decimal import Decimal
from oscar.apps.partner import prices


TaxationDetail = namedtuple(
    'TaxationDetail', ['authority_name', 'tax_name', 'tax_applied', 'fee_applied'])


def _monkey_add_tax(self, authority_name, tax_name, tax_applied='0.00', fee_applied='0.00'):
    tax_applied = Decimal(tax_applied)
    fee_applied = Decimal(fee_applied)

    taxation_detail = TaxationDetail(
        authority_name=authority_name,
        tax_name=tax_name,
        tax_applied=tax_applied,
        fee_applied=fee_applied)
    self.taxation_details.append(taxation_detail)

    if self.tax is None:
        self.tax = Decimal('0.00')
    self.tax += taxation_detail.tax_applied
    self.tax += taxation_detail.fee_applied


def _monkey_clear_taxes(self):
    self.taxation_details = []
    self.tax = None


def monkey_patch_fixed_price():
    """
    MonkeyPatch a few new properties onto oscar.apps.partner.prices.FixedPrice.
    """
    prices.FixedPrice.taxation_details = []
    prices.FixedPrice.add_tax = _monkey_add_tax
    prices.FixedPrice.clear_taxes = _monkey_clear_taxes



from django.db import models, transaction
from django_statsd.clients import statsd
from oscar.core.loading import get_model, get_class
from .settings import CCH_TOLERATE_FAILURE_DURING_PLACE_ORDER as ALLOW_FAIL

OrderTaxation = get_model('cch', 'OrderTaxation')

OrderTotalCalculator = get_class('checkout.calculators', 'OrderTotalCalculator')
CCHTaxCalculator = get_class('oscarcch.calculator', 'CCHTaxCalculator')


class CCHOrderMixin(models.Model):
    is_tax_known = models.BooleanField(default=True)

    class Meta:
        abstract = True


class CCHOrderLineMixin(models.Model):
    basket_line = models.OneToOneField('basket.Line',
        related_name='order_line',
        on_delete=models.SET_NULL,
        null=True)

    class Meta:
        abstract = True


class CCHOrderCreatorMixin(object):
    def place_order(self, basket, total,  # noqa (too complex (12))
                    shipping_method, shipping_charge, user=None,
                    shipping_address=None, billing_address=None,
                    order_number=None, status=None, **kwargs):
        # Calculate the tax liability for this shipping address
        cch_response = CCHTaxCalculator().apply_taxes(basket, shipping_address, ignore_cch_fail=ALLOW_FAIL)
        is_tax_known = (cch_response is not None)
        if not is_tax_known:
            statsd.incr('order.tax_calculation_failed')

        # Update order total now that we now taxes
        total = OrderTotalCalculator().calculate(basket, shipping_charge)

        with transaction.atomic():
            # Save order and line records
            order = super().place_order(
                basket, total,
                shipping_method, shipping_charge, user,
                shipping_address, billing_address,
                order_number, status, is_tax_known=is_tax_known, **kwargs)

            # Save order taxation details
            if is_tax_known:
                OrderTaxation.save_details(order, cch_response)

        statsd.incr('order.placed')
        return order


    def create_line_models(self, order, basket_line, extra_line_fields=None):
        order_line = super().create_line_models(order, basket_line, extra_line_fields)
        order_line.basket_line = basket_line
        order_line.save()
        return order_line



#!/usr/bin/env python
from setuptools import setup, find_packages, Distribution
import codecs
import os.path

# Make sure versiontag exists before going any further
Distribution().fetch_build_eggs('versiontag>=1.2.0')

from versiontag import get_version, cache_git_tag  # NOQA


packages = find_packages('src')

install_requires = [
    'django-oscar>=1.6.0,<2.0.0',
    'django-statsd-mozilla>=0.3.16',
    'instrumented-soap>=1.1.0',
]

extras_require = {
    'development': [
        'psycopg2cffi>=2.7.7',
        'flake8>=3.2.1',
        'freezegun>=0.3.8',
        'lxml>=3.7.2',
        'sphinx>=1.5.2',
        'sphinx-rtd-theme>=0.4.3',
        'tox>=2.6.0',
        'versiontag>=1.2.0',
    ],
}


def fpath(name):
    return os.path.join(os.path.dirname(__file__), name)


def read(fname):
    return codecs.open(fpath(fname), encoding='utf-8').read()


cache_git_tag()

setup(
    name='django-oscar-cch',
    description="Integration between django-oscar and the CCH Sales Tax Office SOAP API.",
    version=get_version(pypi=True),
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    author='Craig Weber',
    author_email='crgwbr@gmail.com',
    url='https://gitlab.com/thelabnyc/django-oscar/django-oscar-cch',
    license='ISC',
    package_dir={'': 'src'},
    packages=packages,
    install_requires=install_requires,
    extras_require=extras_require,
)



from decimal import Decimal as D
from django.test import TestCase
from oscar.core.loading import get_model, get_class
from oscar.test import factories
from soap.tests import SoapTest

Basket = get_model('basket', 'Basket')
ShippingAddress = get_model('order', 'ShippingAddress')
Country = get_model('address', 'Country')
PartnerAddress = get_model('partner', 'PartnerAddress')
USStrategy = get_class('partner.strategy', 'US')


def p(xin):
    """Build a prefix independent XPath"""
    xout = []
    for seg in xin.split('/'):
        xout.append("*[local-name()='%s']" % seg)
    return "/".join(xout)


class BaseTest(SoapTest, TestCase):
    def setUp(self):
        super().setUp()
        Country.objects.create(
            iso_3166_1_a2='US',
            is_shipping_country=True,
            iso_3166_1_a3='USA',
            iso_3166_1_numeric='840',
            display_order=0,
            name="United States of America",
            printable_name="United States")


    def prepare_basket(self, lines=1):
        basket = Basket()
        basket.strategy = USStrategy()

        for i in range(lines):
            product = factories.create_product()
            record = factories.create_stockrecord(
                currency='USD',
                product=product,
                price_excl_tax=D('10.00'))
            factories.create_purchase_info(record)
            basket.add(product)

        from_address = PartnerAddress()
        from_address.line1 = '221 Baker st'
        from_address.line2 = 'B'
        from_address.line4 = 'Anchorage'
        from_address.state = 'AK'
        from_address.postcode = '99501'
        from_address.country = Country.objects.get(pk='US')
        from_address.partner = record.partner
        from_address.save()
        return basket

    def prepare_basket_full_zip(self, lines=1):
        basket = Basket()
        basket.strategy = USStrategy()

        for i in range(lines):
            product = factories.create_product()
            record = factories.create_stockrecord(
                currency='USD',
                product=product,
                price_excl_tax=D('10.00'))
            factories.create_purchase_info(record)
            basket.add(product)

        from_address = PartnerAddress()
        from_address.line1 = '325 F st'
        from_address.line2 = ''
        from_address.line4 = 'Anchorage'
        from_address.state = 'AK'
        from_address.postcode = '99501-2217'
        from_address.country = Country.objects.get(pk='US')
        from_address.partner = record.partner
        from_address.save()
        return basket

    def get_to_address_ohio_short_zip(self):
        to_address = ShippingAddress()
        to_address.line1 = '33001 STATE ROUTE 206'
        to_address.line2 = ''
        to_address.line4 = 'BRINKHAVEN'
        to_address.state = 'OH'
        to_address.postcode = '43006'
        to_address.country = Country.objects.get(pk='US')
        to_address.save()
        return to_address

    def get_to_address_ohio_full_zip(self):
        to_address = ShippingAddress()
        to_address.line1 = '200 HIGH ST'
        to_address.line2 = ''
        to_address.line4 = 'BRINKHAVEN'
        to_address.state = 'OH'
        to_address.postcode = '43006-9000'
        to_address.country = Country.objects.get(pk='US')
        to_address.save()
        return to_address

    def get_to_address(self):
        to_address = ShippingAddress()
        to_address.line1 = '123 Evergreen Terrace'
        to_address.line2 = 'Apt #1'
        to_address.line4 = 'Brooklyn'
        to_address.state = 'NY'
        to_address.postcode = '11201'
        to_address.country = Country.objects.get(pk='US')
        to_address.save()
        return to_address


    def _get_cch_response_normal(self, line_id):
        resp = """
            <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">
                <s:Body>
                    <CalculateRequestResponse xmlns="http://schemas.cch.com/STOService/3.5">
                        <CalculateRequestResult xmlns:a="http://schemas.cch.com/TaxResponse/3.5" xmlns:i="http://www.w3.org/2001/XMLSchema-instance">
                            <a:LineItemTaxes xmlns:b="http://schemas.cch.com/LineItemTax/3.5">
                                <b:LineItemTax>
                                    <b:CountryCode>US</b:CountryCode>
                                    <b:ID>""" + str(line_id) + """</b:ID>
                                    <b:StateOrProvince>NY</b:StateOrProvince>
                                    <b:TaxDetails xmlns:c="http://schemas.cch.com/TaxDetail/3.5">
                                        <c:TaxDetail>
                                            <c:AuthorityName>NEW YORK, STATE OF</c:AuthorityName>
                                            <c:AuthorityType>1</c:AuthorityType>
                                            <c:BaseType>00</c:BaseType>
                                            <c:ExemptAmt>0.000</c:ExemptAmt>
                                            <c:ExemptQty>0.00000000</c:ExemptQty>
                                            <c:FeeApplied>0.000000</c:FeeApplied>
                                            <c:PassFlag>1</c:PassFlag>
                                            <c:PassType>00</c:PassType>
                                            <c:TaxApplied>0.40</c:TaxApplied>
                                            <c:TaxName>STATE SALES TAX-GENERAL MERCHANDISE</c:TaxName>
                                            <c:TaxRate>0.040000000000</c:TaxRate>
                                            <c:TaxableAmount>10.000</c:TaxableAmount>
                                            <c:TaxableQuantity>0.00000000</c:TaxableQuantity>
                                        </c:TaxDetail>
                                        <c:TaxDetail>
                                            <c:AuthorityName>NEW YORK, CITY OF</c:AuthorityName>
                                            <c:AuthorityType>3</c:AuthorityType>
                                            <c:BaseType>00</c:BaseType>
                                            <c:ExemptAmt>0.000</c:ExemptAmt>
                                            <c:ExemptQty>0.00000000</c:ExemptQty>
                                            <c:FeeApplied>0.000000</c:FeeApplied>
                                            <c:PassFlag>1</c:PassFlag>
                                            <c:PassType>00</c:PassType>
                                            <c:TaxApplied>0.45</c:TaxApplied>
                                            <c:TaxName>COUNTY SALES TAX-GENERAL MERCHANDISE</c:TaxName>
                                            <c:TaxRate>0.045000000000</c:TaxRate>
                                            <c:TaxableAmount>10.000</c:TaxableAmount>
                                            <c:TaxableQuantity>0.00000000</c:TaxableQuantity>
                                        </c:TaxDetail>
                                        <c:TaxDetail>
                                            <c:AuthorityName>METROPOLITAN TRANSPORTATION AUTHORITY</c:AuthorityName>
                                            <c:AuthorityType>4</c:AuthorityType>
                                            <c:BaseType>00</c:BaseType>
                                            <c:ExemptAmt>0.000</c:ExemptAmt>
                                            <c:ExemptQty>0.00000000</c:ExemptQty>
                                            <c:FeeApplied>0.000000</c:FeeApplied>
                                            <c:PassFlag>1</c:PassFlag>
                                            <c:PassType>00</c:PassType>
                                            <c:TaxApplied>0.04</c:TaxApplied>
                                            <c:TaxName>COUNTY LOCAL SALES TAX-GENERAL MERCHANDISE</c:TaxName>
                                            <c:TaxRate>0.003750000000</c:TaxRate>
                                            <c:TaxableAmount>10.000</c:TaxableAmount>
                                            <c:TaxableQuantity>0.00000000</c:TaxableQuantity>
                                        </c:TaxDetail>
                                    </b:TaxDetails>
                                    <b:TotalTaxApplied>0.89</b:TotalTaxApplied>
                                </b:LineItemTax>
                            </a:LineItemTaxes>
                            <a:Messages i:nil="true" xmlns:b="http://schemas.cch.com/Message/3.5"/>
                            <a:TotalTaxApplied>0.89</a:TotalTaxApplied>
                            <a:TransactionID>40043</a:TransactionID>
                            <a:TransactionStatus>4</a:TransactionStatus>
                        </CalculateRequestResult>
                    </CalculateRequestResponse>
                </s:Body>
            </s:Envelope>"""
        return resp.encode('utf8')

    def _get_cch_response_ohio_request_short_zip(self, line_id):
        resp = """
        <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">
            <s:Body>
                <CalculateRequestResponse xmlns="http://schemas.cch.com/STOService/3.5">
                    <CalculateRequestResult xmlns:a="http://schemas.cch.com/TaxResponse/3.5" xmlns:i="http://www.w3.org/2001/XMLSchema-instance">
                        <a:LineItemTaxes xmlns:b="http://schemas.cch.com/LineItemTax/3.5">
                            <b:LineItemTax>
                            <b:CountryCode>US</b:CountryCode>
                            <b:ID>""" + str(line_id) + """</b:ID>
                            <b:StateOrProvince>OH</b:StateOrProvince>
                            <b:TaxDetails xmlns:c="http://schemas.cch.com/TaxDetail/3.5">
                                <c:TaxDetail>
                                <c:AuthorityName>OHIO, STATE OF</c:AuthorityName>
                                <c:AuthorityType>1</c:AuthorityType>
                                <c:BaseType>00</c:BaseType>
                                <c:ExemptAmt>0.000</c:ExemptAmt>
                                <c:ExemptQty>0.00000000</c:ExemptQty>
                                <c:FeeApplied>0.000000</c:FeeApplied>
                                <c:PassFlag>1</c:PassFlag>
                                <c:PassType>00</c:PassType>
                                <c:TaxApplied>0.58</c:TaxApplied>
                                <c:TaxName>STATE SALES TAX-GENERAL MERCHANDISE</c:TaxName>
                                <c:TaxRate>0.057500000000</c:TaxRate>
                                <c:TaxableAmount>10.000</c:TaxableAmount>
                                <c:TaxableQuantity>0.00000000</c:TaxableQuantity>
                                </c:TaxDetail>
                                <c:TaxDetail>
                                <c:AuthorityName>KNOX, COUNTY OF</c:AuthorityName>
                                <c:AuthorityType>2</c:AuthorityType>
                                <c:BaseType>00</c:BaseType>
                                <c:ExemptAmt>0.000</c:ExemptAmt>
                                <c:ExemptQty>0.00000000</c:ExemptQty>
                                <c:FeeApplied>0.000000</c:FeeApplied>
                                <c:PassFlag>1</c:PassFlag>
                                <c:PassType>00</c:PassType>
                                <c:TaxApplied>0.10</c:TaxApplied>
                                <c:TaxName>COUNTY SALES TAX-GENERAL MERCHANDISE</c:TaxName>
                                <c:TaxRate>0.010000000000</c:TaxRate>
                                <c:TaxableAmount>10.000</c:TaxableAmount>
                                <c:TaxableQuantity>0.00000000</c:TaxableQuantity>
                                </c:TaxDetail>
                            </b:TaxDetails>
                            <b:TotalTaxApplied>0.68</b:TotalTaxApplied>
                        </b:LineItemTax>
                    </a:LineItemTaxes>
                    <a:Messages i:nil="true" xmlns:b="http://schemas.cch.com/Message/3.5"/>
                    <a:TotalTaxApplied>0.68</a:TotalTaxApplied>
                    <a:TransactionID>121424</a:TransactionID>
                    <a:TransactionStatus>4</a:TransactionStatus>
                </CalculateRequestResult>
            </CalculateRequestResponse>
            </s:Body>
        </s:Envelope>
        """
        return resp.encode('utf8')

    def _get_cch_response_ohio_request_full_zip(self, line_id):
        resp = """
        <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">
            <s:Body>
                <CalculateRequestResponse xmlns="http://schemas.cch.com/STOService/3.5">
                    <CalculateRequestResult xmlns:a="http://schemas.cch.com/TaxResponse/3.5" xmlns:i="http://www.w3.org/2001/XMLSchema-instance">
                        <a:LineItemTaxes xmlns:b="http://schemas.cch.com/LineItemTax/3.5">
                            <b:LineItemTax>
                                <b:CountryCode>US</b:CountryCode>
                                <b:ID>""" + str(line_id) + """</b:ID>
                                <b:StateOrProvince>OH</b:StateOrProvince>
                                <b:TaxDetails xmlns:c="http://schemas.cch.com/TaxDetail/3.5">
                                    <c:TaxDetail>
                                        <c:AuthorityName>OHIO, STATE OF</c:AuthorityName>
                                        <c:AuthorityType>1</c:AuthorityType>
                                        <c:BaseType>00</c:BaseType>
                                        <c:ExemptAmt>0.000</c:ExemptAmt>
                                        <c:ExemptQty>0.00000000</c:ExemptQty>
                                        <c:FeeApplied>0.000000</c:FeeApplied>
                                        <c:PassFlag>1</c:PassFlag>
                                        <c:PassType>00</c:PassType>
                                        <c:TaxApplied>0.58</c:TaxApplied>
                                        <c:TaxName>STATE SALES TAX-GENERAL MERCHANDISE</c:TaxName>
                                        <c:TaxRate>0.057500000000</c:TaxRate>
                                        <c:TaxableAmount>10.000</c:TaxableAmount>
                                        <c:TaxableQuantity>0.00000000</c:TaxableQuantity>
                                    </c:TaxDetail><c:TaxDetail>
                                        <c:AuthorityName>COSHOCTON, COUNTY OF</c:AuthorityName>
                                        <c:AuthorityType>2</c:AuthorityType>
                                        <c:BaseType>00</c:BaseType>
                                        <c:ExemptAmt>0.000</c:ExemptAmt>
                                        <c:ExemptQty>0.00000000</c:ExemptQty>
                                        <c:FeeApplied>0.000000</c:FeeApplied>
                                        <c:PassFlag>1</c:PassFlag>
                                        <c:PassType>00</c:PassType>
                                        <c:TaxApplied>0.15</c:TaxApplied>
                                        <c:TaxName>COUNTY SALES TAX-GENERAL MERCHANDISE</c:TaxName>
                                        <c:TaxRate>0.015000000000</c:TaxRate>
                                        <c:TaxableAmount>10.000</c:TaxableAmount>
                                        <c:TaxableQuantity>0.00000000</c:TaxableQuantity>
                                    </c:TaxDetail>
                                </b:TaxDetails>
                                <b:TotalTaxApplied>0.73</b:TotalTaxApplied>
                            </b:LineItemTax>
                        </a:LineItemTaxes>
                        <a:Messages i:nil="true" xmlns:b="http://schemas.cch.com/Message/3.5"/>
                        <a:TotalTaxApplied>0.73</a:TotalTaxApplied>
                        <a:TransactionID>121425</a:TransactionID>
                        <a:TransactionStatus>4</a:TransactionStatus>
                    </CalculateRequestResult>
                </CalculateRequestResponse>
            </s:Body>
        </s:Envelope>
        """
        return resp.encode('utf8')

    def _get_cch_response_empty(self):
        resp = """
            <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">
                <s:Body>
                    <CalculateRequestResponse xmlns="http://schemas.cch.com/STOService/3.5">
                        <CalculateRequestResult xmlns:a="http://schemas.cch.com/TaxResponse/3.5" xmlns:i="http://www.w3.org/2001/XMLSchema-instance">
                            <a:LineItemTaxes xmlns:b="http://schemas.cch.com/LineItemTax/3.5"/>
                            <a:Messages i:nil="true" xmlns:b="http://schemas.cch.com/Message/3.5"/>
                            <a:TotalTaxApplied>0.00</a:TotalTaxApplied>
                            <a:TransactionID>40043</a:TransactionID>
                            <a:TransactionStatus>4</a:TransactionStatus>
                        </CalculateRequestResult>
                    </CalculateRequestResponse>
                </s:Body>
            </s:Envelope>"""
        return resp.encode('utf8')



from decimal import Decimal
from django.conf import settings
import pytz


def overridable(name, default=None, required=False):
    if required:
        if not hasattr(settings, name) or not getattr(settings, name):
            raise AttributeError("Attribute %s must be defined in Django settings" % name)
    return getattr(settings, name, default)


#: Full URL of the CCH WSDL.
CCH_WSDL = overridable('CCH_WSDL', required=True)

#: Max number of times to retry to calculate tax before giving up.
CCH_MAX_RETRIES = overridable('CCH_MAX_RETRIES', 2)

#: Default entity code to send to CCH.
CCH_ENTITY = overridable('CCH_ENTITY', required=True)

#: Default division code to send to CCH.
CCH_DIVISION = overridable('CCH_DIVISION', required=True)

#: Name of the source system to send to CCH. Defaults to `Oscar`.
CCH_SOURCE_SYSTEM = overridable('CCH_SOURCE_SYSTEM', 'Oscar')

#: Whether or not to set the test flag in CCH requests. Defaults to the same value as Django's ``DEBUG`` setting.
CCH_TEST_TRANSACTIONS = overridable('CCH_TEST_TRANSACTIONS', settings.DEBUG)

#: CCH Transaction Type Code. Defaults to ``01``.
CCH_TRANSACTION_TYPE = overridable('CCH_TRANSACTION_TYPE', '01')

#: CCH Customer Type. Defaults to ``08``.
CCH_CUSTOMER_TYPE = overridable('CCH_CUSTOMER_TYPE', '08')

#: CCH Provider Type. Defaults to ``70``.
CCH_PROVIDER_TYPE = overridable('CCH_PROVIDER_TYPE', '70')

#: Whether or not to set the CCH finalize transaction flag. Defaults to False.
CCH_FINALIZE_TRANSACTION = overridable('CCH_FINALIZE_TRANSACTION', False)

#: Default CCH Product SKU. Can be overridden by creating and setting a Product attribute called cch_product_sku.
CCH_PRODUCT_SKU = overridable('CCH_PRODUCT_SKU', '')

#: Default CCH Product Group Code. Can be overridden by creating and setting a Product attribute called cch_product_group.
CCH_PRODUCT_GROUP = overridable('CCH_PRODUCT_GROUP', '')

#: Default CCH Product Item Code. Can be overridden by creating and setting a Product attribute called cch_product_item.
CCH_PRODUCT_ITEM = overridable('CCH_PRODUCT_ITEM', '')

#: When using the :ref:`usage_simple_integration`, this controls whether or not to allow placing an order when the call to
#: CCH for tax calculation fails or times out. Defaults to ``True``. When ``False`` and an error occurs,
#: OrderCreator.place_order will raise an Exception.
CCH_TOLERATE_FAILURE_DURING_PLACE_ORDER = overridable('CCH_TOLERATE_FAILURE_DURING_PLACE_ORDER', True)

#: Decimal precision to use when sending prices to CCH. Defaults to two-decimal places.
CCH_PRECISION = overridable('CCH_PRECISION', Decimal('.01'))

#: Max length of postal-codes to send to CCH. Defaults to ``5``. All digits and characters after this limit will
#: be clipped in the SOAP request.
CCH_POSTALCODE_LENGTH = overridable('CCH_POSTALCODE_LENGTH', 5)

#: Timezone to use for date times sent to CCH. Defaults to ``UTC``.
CCH_TIME_ZONE = pytz.timezone( overridable('CCH_TIME_ZONE', 'UTC') )



from decimal import Decimal as D
from oscar.core.loading import get_model, get_class
from .base import BaseTest
import unittest


Basket = get_model('basket', 'Basket')
ShippingAddress = get_model('order', 'ShippingAddress')
Country = get_model('address', 'Country')
PartnerAddress = get_model('partner', 'PartnerAddress')
Range = get_model('offer', 'Range')
Benefit = get_model('offer', 'Benefit')
Condition = get_model('offer', 'Condition')
ConditionalOffer = get_model('offer', 'ConditionalOffer')

USStrategy = get_class('partner.strategy', 'US')
Applicator = get_class('offer.applicator', 'Applicator')
CCHTaxCalculator = get_class('oscarcch.calculator', 'CCHTaxCalculator')


@unittest.skip("Disabled because it uses real cch")
class CCHTaxCalculatorRealTest(BaseTest):


    def test_apply_taxes_five_digits_postal_code(self):
        basket = self.prepare_basket_full_zip()
        to_address = self.get_to_address_ohio_short_zip()

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.68'))
        self.assertEqual(basket.total_tax, D('0.68'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.68'))
        self.assertEqual(purchase_info.price.tax, D('0.68'))

        details = purchase_info.price.taxation_details
        self.assertEqual(len(details), 2)
        self.assertEqual(details[0].authority_name, 'OHIO, STATE OF')
        self.assertEqual(details[0].tax_name, 'STATE SALES TAX-GENERAL MERCHANDISE')
        self.assertEqual(details[0].tax_applied, D('0.58'))
        self.assertEqual(details[0].fee_applied, D('0.00'))

    def test_apply_taxes_nine_digits_postal_code(self):
        basket = self.prepare_basket_full_zip()
        to_address = self.get_to_address_ohio_full_zip()

        CCHTaxCalculator().apply_taxes(basket, to_address)

        self.assertTrue(basket.is_tax_known)
        print("basket: %s" % basket)
        self.assertEqual(basket.total_excl_tax, D('10.00'))
        self.assertEqual(basket.total_incl_tax, D('10.73'))
        self.assertEqual(basket.total_tax, D('0.73'))

        purchase_info = basket.all_lines()[0].purchase_info
        self.assertEqual(purchase_info.price.excl_tax, D('10.00'))
        self.assertEqual(purchase_info.price.incl_tax, D('10.73'))
        self.assertEqual(purchase_info.price.tax, D('0.73'))

        details = purchase_info.price.taxation_details
        self.assertEqual(len(details), 2)
        self.assertEqual(details[0].authority_name, 'OHIO, STATE OF')
        self.assertEqual(details[0].tax_name, 'STATE SALES TAX-GENERAL MERCHANDISE')
        self.assertEqual(details[0].tax_applied, D('0.58'))
        self.assertEqual(details[0].fee_applied, D('0.00'))
